# Generated by CodiumAI

import numpy as np
import re

import wizard
from wizard import DataCube

import pytest

def create_test_cube(shape=(3, 4, 4)):
    cube = np.random.rand(*shape)  # Generate a random cube
    return DataCube(cube=cube, wavelengths=np.arange(shape[0]))


class TestDataCube:

    #  DataCube can be initialized with cube, wavelengths, name, and notation.
    def test_initialize_with_parameters(self):
        cube = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        wavelengths = np.array([400, 500])
        name = "Cube1"
        notation = "nm"
    
        data_cube = DataCube(cube=cube, wavelengths=wavelengths, name=name, notation=notation)
    
        assert data_cube.cube is not None
        assert data_cube.wavelengths is not None
        assert data_cube.name == name
        assert data_cube.notation == notation

    #  DataCube can be concatenated with another DataCube using the '+' operator.
    def test_concatenate_datacubes(self):
        cube1 = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        wavelengths1 = np.array([400, 500])
        name1 = "Cube1"
        notation1 = "A"
    
        cube2 = np.array([[[9, 10], [11, 12]], [[13, 14], [15, 16]]])
        wavelengths2 = np.array([600, 700])
        name2 = "Cube2"
        notation2 = "B"
    
        data_cube1 = DataCube(cube=cube1, wavelengths=wavelengths1, name=name1, notation=notation1)
        data_cube2 = DataCube(cube=cube2, wavelengths=wavelengths2, name=name2, notation=notation2)
    
        result = data_cube1 + data_cube2
    
        assert result.cube is not None
        assert result.wavelengths is not None
        assert result.name == name1
        assert result.notation == notation1

    #  DataCube's shape can be accessed.
    def test_access_shape(self):
        cube = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        data_cube = DataCube(cube=cube)
    
        shape = data_cube.shape
    
        assert shape == (2, 2, 2)

    #  DataCube's length can be accessed.
    def test_access_length(self):
        cube = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        data_cube = DataCube(cube=cube)
    
        length = len(data_cube)
    
        assert length == 2

    #  DataCube's layers can be accessed using the [] operator.
    def test_access_layers(self):
        cube = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        data_cube = DataCube(cube=cube)
    
        layer = data_cube[0]
    
        assert np.array_equal(layer, np.array([[1, 2], [3, 4]]))

    #  DataCube can be initialized with no parameters.
    def test_initialize_with_no_parameters(self):
        data_cube = DataCube()
    
        assert data_cube.cube is None
        assert data_cube.wavelengths is None
        assert data_cube.name is None
        assert data_cube.notation is None

    #  DataCube can be initialized with a cube but no wavelengths.
    def test_initialize_with_cube_no_wavelengths(self):
        cube = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        data_cube = DataCube(cube=cube)
    
        assert data_cube.cube is not None
        assert data_cube.wavelengths is not None

    #  DataCube can be concatenated with another DataCube with different x and y shapes.
    def test_concatenate_datacubes_different_shapes(self):
        cube1 = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        wavelengths1 = np.array([400, 500])
        name1 = "Cube1"
        notation1 = "A"
    
        cube2 = np.array([[[9, 10, 11], [12, 13, 14]], [[15, 16, 17], [18, 19, 20]]])
        wavelengths2 = np.array([600, 700])
        name2 = "Cube2"
        notation2 = "B"
    
        data_cube1 = DataCube(cube=cube1, wavelengths=wavelengths1, name=name1, notation=notation1)
        data_cube2 = DataCube(cube=cube2, wavelengths=wavelengths2, name=name2, notation=notation2)
    
        with pytest.raises(ValueError):
            result = data_cube1 + data_cube2

    # Initializes a DataCube instance with all parameters.
    def test_all_parameters(self):
        cube = np.zeros((10, 100, 100))
        wavelengths = [400, 410, 420, 430, 440, 450, 460, 470, 480, 490]
        name = "Example DataCube"
        notation = "wavelengths"
        record = True

        data_cube = DataCube(cube=cube, wavelengths=wavelengths, name=name, notation=notation, record=record)

        assert data_cube.name == name
        assert data_cube.shape == cube.shape
        assert data_cube.dim is None
        assert np.array_equal(data_cube.wavelengths, np.array(wavelengths))
        assert np.array_equal(data_cube.cube, cube)
        assert data_cube.notation == notation
        assert data_cube.record is True

    # Add a datacube and an int and check if the ValueError gets raised.
    def test_add_with_int_raises_valueerror(self):
        cube1 = DataCube(cube=np.ones((3, 2, 2)))
        cube2 = 5
        with pytest.raises(ValueError):
            cube1 + cube2

    # Adding two DataCubes with same shape and one of them has empty wavelengths should return a new DataCube with
    # concatenated cube and None wavelengths, and a warning message
    def test_same_shape_one_empty_wavelengths(self):
        dc1 = DataCube(cube=np.ones((10, 100, 100)), wavelengths=np.arange(10), name="dc1")
        dc2 = DataCube(cube=np.ones((10, 100, 100)), wavelengths=None, name="dc2")
        dc2.wavelengths = None

        with pytest.warns(UserWarning):
            result = dc1 + dc2

        assert isinstance(result, DataCube)
        assert result.cube.shape == (20, 100, 100)
        assert result.cube.shape[0] == len(result.wavelengths)

    #  Set a valid 3D numpy array to the cube attribute
    def test_set_valid_3d_array(self):
        cube_data = np.array([[[1, 2, 3], [4, 5, 6], [7, 8, 9]]])
        dc = DataCube()
        dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, cube_data)

    #  Set a valid 4D numpy array to the cube attribute
    def test_set_valid_4d_array(self):
        cube_data = np.array([[[[1, 2, 3], [4, 5, 6], [7, 8, 9]]]])
        dc = DataCube()
        dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, cube_data)

    #  Set a valid 2D numpy array to the cube attribute
    def test_set_valid_2d_array(self):
        cube_data = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        dc = DataCube()
        dc.set_cube(cube_data)
        expected_cube = np.zeros(shape=(1, cube_data.shape[0], cube_data.shape[1]), dtype=cube_data.dtype)
        expected_cube[0] = cube_data
        assert np.array_equal(dc.cube, expected_cube)

    #  Set a valid 3D numpy array to the cube attribute with the same shape as an existing cube
    def test_set_valid_3d_array_same_shape(self):
        existing_cube = np.array([[[1, 2, 3], [4, 5, 6], [7, 8, 9]]])
        dc = DataCube(cube=existing_cube)
        cube_data = np.array([[[10, 11, 12], [13, 14, 15], [16, 17, 18]]])
        dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, cube_data)

    #  Set a valid 4D numpy array to the cube attribute with the same shape as an existing cube
    def test_set_valid_4d_array_same_shape(self):
        existing_cube = np.array([[[[1, 2, 3], [4, 5, 6], [7, 8, 9]]]])
        dc = DataCube(cube=existing_cube)
        cube_data = np.array([[[[10, 11, 12], [13, 14, 15], [16, 17, 18]]]])
        dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, cube_data)

    #  Set an empty numpy array to the cube attribute
    def test_set_empty_array(self):
        cube_data = np.array([])
        dc = DataCube()
        with pytest.raises(AttributeError):
            dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, None)

    #  Set a list to the cube attribute
    def test_set_list(self):
        cube_data = [[[1, 2, 3], [4, 5, 6], [7, 8, 9]]]
        dc = DataCube()
        dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, np.array(cube_data))

    # test __iter__ and __next__
    def test_returns_next_item_and_wavelength(self):
        dc = DataCube(cube=np.random.rand(10, 100, 100), wavelengths=np.arange(10))
        iterator = iter(dc)
        next_value = next(iterator)
        assert isinstance(next_value, tuple)
        assert len(next_value) == 2
        assert isinstance(next_value[0], np.ndarray)

    # Should return a string with the name and shape of the DataCube
    def test_return_string_with_name_and_shape(self):
        dc = DataCube(cube=np.ones((3, 4, 5)), wavelengths=[400, 500, 600], name="Example", notation='nm')
        expected_output = ("Name: Example\n"
                           "Shape: (3, 4, 5)\n"
                           "Wavelengths:\n"
                           "\tLen: 3\n"
                           "\tFrom: 400\n"
                           "\tTo: 600\n"
                           "Notaion: nm"
                          )

        assert dc.__str__() == expected_output



    #  Method sets wavelength data when given a valid 1D numpy array.
    def test_valid_wavelengths_array(self):
        wavelengths = np.array([400, 450, 500, 550, 600])
        dc = DataCube()
        dc.set_wavelengths(wavelengths)
        assert np.array_equal(dc.wavelengths, wavelengths)

    #  Method sets wavelength data when given an empty 1D numpy array.
    def test_empty_wavelengths_array(self):
        wavelengths = np.array([])
        dc = DataCube()
        dc.set_wavelengths(wavelengths)
        assert np.array_equal(dc.wavelengths, wavelengths)

    #  Method raises AttributeError when given a non-convertible object.
    def test_non_convertible_object(self):
        wavelengths = "invalid"
        dc = DataCube()
        with pytest.raises(AttributeError):
            dc.set_wavelengths(wavelengths)

    #  Method raises AttributeError when given a numpy array with ndim != 1.
    def test_ndim_not_1(self):
        wavelengths = np.array([[400, 450, 500], [550, 600, 650]])
        dc = DataCube()
        with pytest.raises(AttributeError):
            dc.set_wavelengths(wavelengths)

    #  Method raises AttributeError when given a numpy array with ndim == 0.
    def test_ndim_0(self):
        wavelengths = np.array(400)
        dc = DataCube()
        with pytest.raises(AttributeError):
            dc.set_wavelengths(wavelengths)

    # tets add function
    def test_add_with_none(self):
        dc1 = DataCube()
        dc2 = DataCube()
        with pytest.raises(ValueError):
            dc3 = dc1 + dc2

    # test set method
    def test_setitem(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        a = np.random.rand(100, 100)
        dc[3] = a
        assert np.array_equal(dc.cube[3], a)

    # Test itter method
    def test_next(self):
        idx = 0
        dc = DataCube(cube=np.random.rand(5,10,10))
        for idx, i in enumerate(dc):
            pass
        assert dc.shape[0] == idx + 1

    # Resizing a cube with valid x_new and y_new values and linear interpolation
    def test_resize_linear_down(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        dc.resize(50, 50, interpolation='linear')
        assert dc.cube.shape == (10, 50, 50)

    # Resizing a cube with valid x_new and y_new values and nearest interpolation
    def test_valid_resize_nearest_interpolation(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        dc.resize(50, 50, interpolation='nearest')
        assert dc.cube.shape == (10, 50, 50)

    # Resizing a cube with x_new and y_new values
    def test_resize_linear(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        dc.resize(200, 200, interpolation='linear')
        assert dc.cube.shape == (10, 200, 200)

    # Resizing a cube with x_new and y_new values
    def test_resize_nearest(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        dc.resize(200, 200, interpolation='nearest')
        assert dc.cube.shape == (10, 200, 200)

    # Resizing a cube with x_new and y_new values
    def test_resize_area(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        dc.resize(200, 200, interpolation='area')
        assert dc.cube.shape == (10, 200, 200)

    # Resizing a cube with x_new and y_new values
    def test_resize_cubic(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        dc.resize(200, 200, interpolation='cubic')
        assert dc.cube.shape == (10, 200, 200)

    # Resizing a cube with x_new and y_new values
    def test_resize_lanczos(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        dc.resize(200, 200, interpolation='lanczos')
        assert dc.cube.shape == (10, 200, 200)

    def test_resize_unknow_interpolatio_method(self):
        dc = DataCube(cube=np.random.rand(10,10,10))
        with pytest.raises(ValueError):
            dc.resize(200,200, interpolation='test')

    def test_set_wavelengths_with_non_array(self):
        dc = DataCube(cube=np.random.rand(10,10,10))
        dc.set_wavelengths([0,1,2,3,4,5,6,7,8,9])

    def test_set_cube_with_non_array_and_get_errer(self):
        dc = DataCube()
        with pytest.raises(AttributeError):
            dc.set_cube('hi')


    # Line 242: Test set_name with valid and invalid inputs
    def test_set_name(self):
        data_cube = DataCube()
        valid_name = "ValidName"
        data_cube.set_name(valid_name)
        assert data_cube.name == valid_name

        with pytest.raises(AttributeError):
            data_cube.set_name(123)  # Non-string input

    # Lines 261-273: Test set_wavelengths with valid and invalid data
    def test_set_wavelengths(self):
        data_cube = DataCube()

        # Valid 1D array
        valid_wavelengths = np.array([400, 500, 600])
        data_cube.set_wavelengths(valid_wavelengths)
        assert np.array_equal(data_cube.wavelengths, valid_wavelengths)

        # Valid list
        valid_wavelengths_list = [400, 500, 600]
        data_cube.set_wavelengths(valid_wavelengths_list)
        assert np.array_equal(data_cube.wavelengths, np.array(valid_wavelengths_list))

        # Invalid 2D array
        invalid_wavelengths = np.array([[400, 500], [600, 700]])
        with pytest.raises(AttributeError):
            data_cube.set_wavelengths(invalid_wavelengths)

    # Line 280: Test set_cube with valid 3D and 2D arrays, and invalid input
    def test_set_cube(self):
        data_cube = DataCube()

        # Valid 3D array
        valid_cube_3d = np.random.rand(5, 10, 10)
        data_cube.set_cube(valid_cube_3d)
        assert data_cube.cube.shape == (5, 10, 10)

        # Valid 2D array which should be converted to 3D
        valid_cube_2d = np.random.rand(10, 10)
        data_cube.set_cube(valid_cube_2d)
        assert data_cube.cube.shape == (1, 10, 10)

        # Invalid array (not 2D, 3D, or 4D)
        invalid_cube = np.random.rand(1, 10, 10, 10, 10)
        with pytest.raises(AttributeError):
            data_cube.set_cube(invalid_cube)

    # Lines 288-291: Test _set_cube_shape method
    def test_set_cube_shape(self):
        data_cube = DataCube()
        cube = np.random.rand(3, 10, 10)
        data_cube.set_cube(cube)
        assert data_cube.shape == (3, 10, 10)

    # Lines 341-342: Test set_notation with valid input
    def test_set_notation(self):
        data_cube = DataCube()
        notation = "nm"
        data_cube.set_notation(notation)
        assert data_cube.notation == notation

    # Line 370: Test start_recording method
    def test_start_recording(self, mocker):
        data_cube = DataCube()
        mocker.patch("wizard._utils.tracker.TrackExecutionMeta.start_recording")
        data_cube.start_recording()
        assert data_cube.record is True

    # Lines 379-380: Test stop_recording method
    def test_stop_recording(self, mocker):
        data_cube = DataCube()
        mocker.patch("wizard._utils.tracker.TrackExecutionMeta.stop_recording")
        data_cube.stop_recording()
        assert data_cube.record is False

    # Lines 397-410: Test save_template with valid and invalid filenames
    def test_save_template(self, mocker):
        data_cube = DataCube()
        mock_yaml_dump = mocker.patch("yaml.dump")
        mock_open = mocker.patch("builtins.open", mocker.mock_open())

        # Valid filename
        filename = "template"
        data_cube.save_template(filename)
        mock_open.assert_called_with(filename+'.yaml', 'w')
        assert mock_yaml_dump.called

        # Invalid filename: None
        with pytest.raises(AttributeError):
            data_cube.save_template(None)

        # Invalid filename: non-string type
        with pytest.raises(AttributeError):
            data_cube.save_template(123)

    # Lines 431-440: Test _map_args_to_kwargs with sample function and args
    def test_map_args_to_kwargs(self):
        data_cube = DataCube()

        # Define a sample function to test argument mapping
        def sample_func(a, b, c=None):
            pass

        # Test positional and keyword argument mapping
        args = (1, 2)
        kwargs = {"c": 3}
        mapped_kwargs = data_cube._map_args_to_kwargs(sample_func, args, kwargs)
        assert mapped_kwargs == {"a": 1, "b": 2, "c": 3}

    # Lines 491-497: Test execute_template with mocked method
    def test_execute_template(self, mocker):
        data_cube = DataCube()
        mock_method = mocker.patch.object(data_cube, "set_name")

        template_data = [{"method": "set_name", "kwargs": {"name": "Cube1"}}]

        # Mock yaml.safe_load to return template_data
        mocker.patch("yaml.safe_load", return_value=template_data)
        mock_open = mocker.patch("builtins.open", mocker.mock_open(read_data="template_data"))

        data_cube.execute_template("template.yml")
        mock_open.assert_called_with("template.yml", 'rb')
        mock_method.assert_called_once_with(name="Cube1")

    def test_custom_read(self):
        # Create a DataCube instance
        data_cube = DataCube()

        # Define a mock reading function
        def mock_reading_func(*args, **kwargs):
            cube_data = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
            wavelengths = [400, 500]
            notation = "nm"
            name = "TestCube"
            return DataCube(cube=cube_data, wavelengths=wavelengths, name=name, notation=notation)

        # Set the custom reading function
        data_cube.set_custom_reader(mock_reading_func)

        # Invoke custom_read, which should use mock_reading_func
        data_cube.custom_read()

        # Assertions to verify that attributes were set correctly
        assert np.array_equal(data_cube.cube, np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]))
        assert np.array_equal(data_cube.wavelengths, np.array([400, 500]))
        assert data_cube.notation == "nm"
        assert data_cube.name == "TestCube"

        # Remove the reading_func and check for NotImplementedError
        del data_cube.reading_func
        error_message = "You need to implement a `custom_read` method and set it with `set_custom_read(fuc)."
        with pytest.raises(NotImplementedError, match=re.escape(error_message)):
            data_cube.custom_read()

    def test_set_cube_with_wrong_cube_data(self):

        dc = DataCube()

        with pytest.raises(ValueError):
            dc.set_cube([[3123,123123], '134'])
 
    def test_record_start_stop(self):
        dc = DataCube(record=True)
        assert dc.record == True
        dc.stop_recording()
        assert dc.record == False
        dc.start_recording()
        assert dc.record == True


class TestDataCubeOps:

    def test_remove_spikes(self):
        dc = create_test_cube()
        dc.cube[0, 1, 1] = 1000  # Insert a spike
        threshold = 10
        dc.remove_spikes(threshold=threshold, window=3)

        assert dc is not None
        assert dc.cube.shape == dc.cube.shape
        assert dc.cube[0, 1, 1] != 1000




    def test_resize(self):
        dc = create_test_cube(shape=(3, 10, 10))
        x_new, y_new = 5, 5

        dc.resize(x_new, y_new, interpolation='linear')

        assert dc.cube.shape[1] == x_new
        assert dc.cube.shape[2] == y_new

    def test_resize_invalid_interpolation(self):
        dc = create_test_cube(shape=(3, 10, 10))
        with pytest.raises(ValueError):
            dc.resize(5, 5, interpolation='invalid_method')

    def test_baseline_als(self):
        dc = create_test_cube()
        dc.baseline_als(lam=1000000, p=0.01, niter=10)
        assert dc is not None
        assert isinstance(dc, DataCube)

    def test_merge_cubes(self):
        dc1 = create_test_cube()
        dc2 = create_test_cube()
        dc2.set_wavelengths([4,5,6])
        dc1.merge_cubes(dc2)
        assert dc1 is not None
        np.testing.assert_array_equal(dc1.wavelengths, [0,1,2,4,5,6])
        np.testing.assert_array_equal(dc1.shape, (6,4,4))
        assert isinstance(dc1, DataCube)

    def test_remove_background_dark(self):
        dc = create_test_cube(shape=(3, 24, 24))

        # Create a temporary cube with specific values for testing
        tmp_cube = dc.cube.copy() * 0.2

        # Define the middle region that should remain unchanged
        x_start, x_end = int(dc.shape[1] * 0.25), int(dc.shape[1] * 0.75)
        y_start, y_end = int(dc.shape[2] * 0.25), int(dc.shape[2] * 0.75)

        # Apply specific values to the middle region using broadcasting
        tmp_cube[0, x_start:x_end, y_start:y_end] = 0.8
        tmp_cube[1, x_start:x_end, y_start:y_end] = 0.8
        tmp_cube[2, x_start:x_end, y_start:y_end] = 0.8

        # Set the modified cube
        dc.set_cube(tmp_cube)

        # Call the method under test
        threshold = 50
        dc.remove_background(threshold=threshold, style='dark')

        # Assertions
        assert dc is not None, "DataCube instance should not be None after removing the background."
        assert dc.cube.shape == tmp_cube.shape, "The shape of the cube should remain unchanged after background removal."
        assert isinstance(dc.cube, np.ndarray), "The cube should be a NumPy array."

        # Validate that the middle region remains unchanged
        middle_region = dc.cube[:, x_start:x_end, y_start:y_end]
        expected_middle = tmp_cube[:, x_start:x_end, y_start:y_end]
        assert np.allclose(middle_region, expected_middle), "The middle region should remain unchanged for 'dark' style."

        # Validate that the outside regions have a mean close to 1
        outside_regions = np.concatenate([
            dc.cube[:, :x_start, :].flatten(),  # Top region
            dc.cube[:, x_end:, :].flatten(),   # Bottom region
            dc.cube[:, :, :y_start].flatten(), # Left region
            dc.cube[:, :, y_end:].flatten()    # Right region
        ])

        assert np.isclose(outside_regions.mean(), 0, atol=1e-1), "The outside regions should have a mean value close to 1 for 'dark' style."

    def test_remove_background_bright(self):
        dc = create_test_cube(shape=(3, 24, 24))

        # Create a temporary cube with specific low values for testing
        tmp_cube = dc.cube.copy() * 0.6 + 0.6

        # Define the middle region that should remain unchanged
        x_start, x_end = int(dc.shape[1] * 0.25), int(dc.shape[1] * 0.75)
        y_start, y_end = int(dc.shape[2] * 0.25), int(dc.shape[2] * 0.75)

        # Apply specific low values to the middle region using broadcasting
        tmp_cube[0, x_start:x_end, y_start:y_end] = 0.2
        tmp_cube[1, x_start:x_end, y_start:y_end] = 0.2
        tmp_cube[2, x_start:x_end, y_start:y_end] = 0.2

        # Set the modified cube
        dc.set_cube(tmp_cube)

        from matplotlib import pyplot as plt

        # Call the method under test
        threshold = 10
        dc.remove_background(threshold=threshold, style='bright')

        # Assertions
        assert dc is not None, "DataCube instance should not be None after removing the background."
        assert dc.cube.shape == tmp_cube.shape, "The shape of the cube should remain unchanged after background removal."
        assert isinstance(dc.cube, np.ndarray), "The cube should be a NumPy array."

        # Validate that the middle region remains unchanged
        middle_region = dc.cube[:, x_start:x_end, y_start:y_end]
        expected_middle = tmp_cube[:, x_start:x_end, y_start:y_end]
        assert np.allclose(middle_region, expected_middle), "The middle region should remain unchanged for 'bright' style."

        # Validate that the outside regions have a mean close to 1
        outside_regions = np.concatenate([
            dc.cube[:, :x_start, :].flatten(),  # Top region
            dc.cube[:, x_end:, :].flatten(),   # Bottom region
            dc.cube[:, :, :y_start].flatten(), # Left region
            dc.cube[:, :, y_end:].flatten()    # Right region
        ])

        assert np.isclose(outside_regions.mean(), 1, atol=.5), "The outside regions should have a mean value close to 1 for 'bright' style."

    def test_remove_background_invalid_style(self):
        dc = create_test_cube()
        threshold = .5

        # Test for invalid style
        with pytest.raises(ValueError):
            dc.remove_background(threshold=threshold, style='invalid')

    def test_inverse(self):
        dc = create_test_cube()
        inverted_dc = dc.inverse()
        assert inverted_dc is not None
        assert np.all(inverted_dc.cube >= 0)  # Inversion should result in non-negative values

    def test_remove_vignetting(self):
        dc = create_test_cube(shape=(4, 200, 200))
        corrected_dc = dc.remove_vignetting() # axis=1, slice_params={"start": 0, "end": 2, "step": 1})
        assert corrected_dc is not None
        assert corrected_dc.cube.shape == dc.cube.shape

    def test_merge_cubes_invalid_shape(self):
        dc1 = create_test_cube(shape=(3, 4, 4))
        dc2 = create_test_cube(shape=(3, 5, 5))
        with pytest.raises(NotImplementedError):
            dc1.merge_cubes(dc2)

    def test_register_layers_simple(self):
        import numpy as np
        dc = create_test_cube(shape=(3, 5, 5))
        # Make all layers identical for easy registration
        base = np.random.rand(5, 5)
        dc.cube = np.stack([base, base, base])
        result = dc.register_layers_simple(max_features=1000, match_percent=0.2)
        assert result is dc
        assert dc.cube.shape == (3, 5, 5)
        assert getattr(dc, 'registered', False) is True

    def test_remove_vignetting_poly_invalid_axis(self):
        dc = create_test_cube()
        with pytest.raises(ValueError):
            dc.remove_vignetting_poly(axis=3)

    def test_remove_vignetting_poly_default(self):
        import numpy as np
        dc = create_test_cube(shape=(2, 72, 72))
        original_shape = dc.cube.shape
        result = dc.remove_vignetting_poly()
        assert result is dc
        assert dc.cube.shape == original_shape
        # After polynomial vignetting removal, dtype should be float32
        assert dc.cube.dtype == np.float32

    def test_normalize(self):
        import numpy as np
        dc = create_test_cube(shape=(2, 4, 4))
        data = np.array([
            [[ 1,  2,  3,  4],
             [ 5,  6,  7,  8],
             [ 9, 10, 11, 12],
             [13, 14, 15, 16]],
            [[16, 15, 14, 13],
             [12, 11, 10,  9],
             [ 8,  7,  6,  5],
             [ 4,  3,  2,  1]]
        ], dtype=float)
        dc.set_cube(data)
        dc.normalize()
        # Each layer should now span [0,1]
        for layer in dc.cube:
            assert np.isclose(layer.min(), 0.0)
            assert np.isclose(layer.max(), 1.0)

    def test_register_layers_best(self):
        dc = create_test_cube(shape=(10, 251, 201))

        # Create a temporary cube with specific values for testing
        tmp_cube = dc.cube.copy() * 0.1

        # Define the middle region that should remain unchanged
        x_start, x_end = int(dc.shape[1] * 0.45), int(dc.shape[1] * 0.65)
        y_start, y_end = int(dc.shape[2] * 0.45), int(dc.shape[2] * 0.65)

        # Apply specific values to the middle region using broadcasting
        for idx, layer in enumerate(range(dc.shape[0])):
            tmp_cube[layer, x_start+idx:x_end+idx, y_start+idx:y_end+idx] = 0.9
        dc.set_cube(tmp_cube)
        result = dc.register_layers_best(ref_layer=2, scale_thresh=2.2, rot_thresh=20.)
        assert result is dc
        assert dc.cube.shape == (10, 251, 201)
        assert getattr(dc, 'registered', False) is True

    def test_remove_vignetting(self):
        import numpy as np
        dc = create_test_cube(shape=(3, 4, 4))
        # constant cube so flat-field correction yields same values
        constant = np.ones((3, 4, 4), dtype=np.uint8) * 10
        dc.set_cube(constant.copy())
        dtype_before = dc.cube.dtype
        result = dc.remove_vignetting(sigma=1)
        assert result is dc
        assert dc.cube.shape == constant.shape
        assert dc.cube.dtype == dtype_before
        # all values should remain equal to original constant
        assert np.all(dc.cube == constant)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

# Add the path to the parent directory of 'wizard'
sys.path.insert(0, os.path.abspath('../../'))
print(f"Using Python interpreter: {sys.executable}")

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'hsi-wizard'
copyright = '2024, Felix Wuehler'
author = 'Felix Wuehler'
release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# Sphinx extensions
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',  # For Google and NumPy style docstrings
    'sphinx.ext.autodoc.typehints',  # For type hints in function signatures
    'sphinx.ext.viewcode'  # To link to source code
]

templates_path = ['_templates']
exclude_patterns = []

# Include type hints in parameter descriptions
autodoc_typehints = "description"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

#  Explicitly set the problematic options in conf.py to their default or desired values
html_permalinks_icon = "¶"  # Default value
jquery_use_sri = False  # Disable SRI for jQuery

# Mock imports for unavailable or optional dependencies
autodoc_mock_imports = ['nrrd', 'rich'] 


import pytest
import numpy as np
from scipy.signal import savgol_filter, butter, filtfilt

import wizard

# Importing spectral processing functions
from wizard._processing.spectral import (
    smooth_savgol,
    smooth_moving_average,
    smooth_butter_lowpass,
    spec_baseline_als,
    calculate_modified_z_score,
    get_ratio_two_specs,
    get_sub_tow_specs,
    signal_to_noise,
    del_leading_zeros,
    del_last_zeros,
)

# Importing clustering functions
from wizard._processing.cluster import (
    quit_low_change_in_clusters,
    discard_clusters,
    update_clusters,
    initial_clusters,
    sort_arrays_by_first,
    split_clusters,
    compute_avg_distance,
    compute_overall_distance,
    merge_clusters,
    compute_pairwise_distances,
    isodata,
)


# Fixtures for sample data
@pytest.fixture
def sample_spectrum():
    """Fixture: Generate a sample spectrum (range of 100 values)."""
    return np.array(range(100))


@pytest.fixture
def sample_baseline_spectrum():
    """Fixture: Generate a sample baseline spectrum (range of 100 values)."""
    return np.array(range(100))


@pytest.fixture
def sample_wavelengths():
    """Fixture: Generate a sample of wavelengths (10 values between 400 and 800)."""
    return np.linspace(400, 800, 10)


# --------------- Spectral Processing Tests ---------------
class TestSpectralProcessing:
    """Test suite for spectral processing functions."""

    def test_smooth_savgol(self, sample_spectrum):
        """Test Savitzky-Golay smoothing."""
        smoothed = smooth_savgol(sample_spectrum, window_length=5, polyorder=2)
        assert len(smoothed) == len(sample_spectrum)

    def test_smooth_moving_average(self, sample_spectrum):
        """Test moving average smoothing."""
        smoothed = smooth_moving_average(sample_spectrum, window_size=3)
        assert len(smoothed) == len(sample_spectrum) - 2

    def test_smooth_butter_lowpass(self, sample_spectrum):
        """Test Butterworth low-pass filter."""
        filtered = smooth_butter_lowpass(sample_spectrum, cutoff=0.1, fs=1, order=3)
        assert len(filtered) == len(sample_spectrum)

    def test_calculate_modified_z_score(self, sample_spectrum):
        """Test calculation of the modified Z-score."""
        modified_z = calculate_modified_z_score(sample_spectrum)
        assert modified_z.shape[0] == sample_spectrum.shape[0]

    def test_get_ratio_two_specs(self, sample_spectrum, sample_wavelengths):
        """Test calculation of the ratio between two specified wavelengths."""
        ratio = get_ratio_two_specs(sample_spectrum, sample_wavelengths, wave_1=450, wave_2=650)
        assert ratio != -1
        assert ratio >= 0

    def test_get_sub_tow_specs(self, sample_spectrum, sample_wavelengths):
        """Test calculation of the difference between two specified wavelengths."""
        diff = get_sub_tow_specs(sample_spectrum, sample_wavelengths, wave_1=450, wave_2=650)
        assert diff != -1

    def test_signal_to_noise(self, sample_spectrum):
        """Test signal-to-noise ratio calculation."""
        snr = signal_to_noise(sample_spectrum)
        assert snr >= 0

    def test_del_leading_zeros(self):
        """Test removal of leading zeros from a spectrum."""
        spectrum = np.array([0, 0, 0, 5, 6, 7])
        result = del_leading_zeros(spectrum, auto_offset=0)
        assert result[0] == 5

    def test_del_last_zeros(self):
        """Test removal of trailing zeros from a spectrum."""
        spectrum = np.array([5, 6, 7, 0, 0, 0])
        result = del_last_zeros(spectrum, auto_offset=0)
        assert result[-1] == 7


# --------------- Edge Case Tests ---------------
class TestEdgeCases:
    """Test suite for edge cases and boundary conditions."""

    def test_smooth_savgol_invalid_input(self):
        """Test Savitzky-Golay smoothing with invalid window length."""
        with pytest.raises(ValueError):
            smooth_savgol(np.array([1, 2, 3]), window_length=4, polyorder=2)

    def test_signal_to_noise_zero_std(self):
        """Test signal-to-noise ratio calculation when standard deviation is zero."""
        spectrum = np.array([1, 1, 1, 1])
        snr = signal_to_noise(spectrum)
        assert snr == 0


# --------------- Clustering Tests ---------------
class TestClustering:
    """Test suite for clustering-related functions."""

    def setup_method(self):
        """Initialize test data for clustering tests."""
        self.centers = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        self.last_centers = np.array([[1.1, 2.1], [2.9, 3.9], [5.1, 6.1]])
        self.img_flat = np.random.rand(100, 2)
        self.img_class_flat = np.random.randint(0, 3, 100)
        self.clusters_list = np.array([0, 1, 2])
        self.theta_o = 0.05
        self.theta_m = 10

    def test_discard_clusters(self):
        """Test discarding clusters based on a threshold."""
        img_class_flat = np.array([0, 1, 1, 2, 2, 2])
        centers = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        clusters_list = np.array([0, 1, 2])
        theta_m = 2

        new_centers, new_clusters_list, k_ = discard_clusters(img_class_flat, centers, clusters_list, theta_m)
        assert new_centers.shape == (1, 2)
        assert new_clusters_list.size == 1
        assert k_ == 3

    def test_update_clusters(self):
        """Test updating cluster centers."""
        new_centers, new_clusters_list, k_ = update_clusters(
            self.img_flat, self.img_class_flat, self.centers, self.clusters_list
        )
        assert new_centers.shape[1] == self.img_flat.shape[1]

    def test_initial_clusters(self):
        """Test initialization of cluster centers."""
        img_flat = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        k_ = 2

        centers = initial_clusters(img_flat, k_, method="linspace")
        assert centers.shape == (2, 2)

        centers = initial_clusters(img_flat, k_, method="random")
        assert centers.shape == (2, 2)

    def test_sort_arrays_by_first(self):
        """Test sorting of arrays by the first column."""
        sorted_centers, sorted_clusters_list = sort_arrays_by_first(self.centers, self.clusters_list)
        assert sorted_centers[0, 0] <= sorted_centers[1, 0] <= sorted_centers[2, 0]

    def test_split_clusters(self):
        """Test splitting of clusters based on thresholds."""
        img_flat = np.random.rand(10, 2)
        img_class_flat = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2, 2])
        centers = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        clusters_list = np.array([0, 1, 2])
        theta_s = 1.0
        theta_m = 2

        new_centers, new_clusters_list, k_ = split_clusters(
            img_flat, img_class_flat, centers, clusters_list, theta_s, theta_m
        )
        assert new_centers.shape[0] >= centers.shape[0]
        assert new_clusters_list.size >= clusters_list.size

    def test_compute_avg_distance(self):
        """Test computation of average distance within clusters."""
        avg_dists, k_ = compute_avg_distance(self.img_flat, self.img_class_flat, self.centers, self.clusters_list)
        assert avg_dists.shape == (self.centers.shape[0],)

    def test_compute_overall_distance(self):
        """Test computation of overall distance within clusters."""
        avg_dists, _ = compute_avg_distance(self.img_flat, self.img_class_flat, self.centers, self.clusters_list)
        d, k_ = compute_overall_distance(self.img_class_flat, avg_dists, self.clusters_list)
        assert isinstance(d, float)

    def test_merge_clusters(self):
        """Test merging of clusters."""
        new_centers, new_clusters_list, k_ = merge_clusters(
            self.img_class_flat, self.centers, self.clusters_list, 2, 2, 3
        )
        assert isinstance(new_centers, np.ndarray)

    def test_compute_pairwise_distances(self):
        """Test computation of pairwise distances between cluster centers."""
        pair_dists = compute_pairwise_distances(self.centers)
        assert isinstance(pair_dists, list)
        assert len(pair_dists) > 0

    def test_isodata(self):
        """Test ISODATA clustering."""
        dc = wizard.DataCube(cube=np.random.rand(20, 8, 9), wavelengths=np.random.randint(0, 200, size=20))
        result = isodata(dc, k=3, it=10)
        assert isinstance(result, np.ndarray)
        assert result.shape == (8, 9)

    def test_quit_low_change_in_clusters(self):
        """Test termination of clustering based on low change in cluster centers."""
        centers = np.array([[1.0, 2.0], [3.0, 4.0]])
        last_centers = np.array([[1.01, 2.01], [3.01, 4.01]])
        theta_o = 0.02
        assert quit_low_change_in_clusters(centers, last_centers, theta_o) is True

        theta_o = 0.001
        assert quit_low_change_in_clusters(centers, last_centers, theta_o) is False


"""
_utils/_loader/tdms.py
========================

.. module:: tdms
   :platform: Unix
   :synopsis: Provides functions to read and write .tdms files.

Module Overview
---------------

This module includes functions for reading .tdms files, specifically designed to extract spectral data
and organize it into a DataCube format.

Functions
---------

.. autofunction:: _read_tdms

"""
import re
from nptdms import TdmsFile

from ._helper import to_cube
from ..._core import DataCube


# Precompile regex for length extraction
def _extract_dims(col_name):
    nums = re.findall(r"\d+", col_name)
    return int(nums[0]) + 1, int(nums[1]) + 1


def _read_tdms(path: str) -> DataCube:
    """
    Optimized TDMS reader: reads the file, cleans and filters columns,
    vectorizes operations, and builds DataCube efficiently.
    """
    # Load and convert to DataFrame
    tdms = TdmsFile(path)
    df = tdms.as_dataframe()

    # Clean column names in-place
    cols = df.columns.str.replace(' ', '', regex=False).str.replace("'", '', regex=False)
    df.columns = cols

    # Identify columns by mask
    is_raw = cols.str.contains('RAW')
    is_drop = cols.str.contains('DarkCurrent') | cols.str.contains('cm|nm')
    is_sample = ~(is_raw | is_drop)

    # Determine data type and wavelength column offset
    if cols.str.contains('RAMAN').any():
        data_type, wave_col = 'raman', 1
    elif cols.str.contains('NIR|KNIR').any():
        data_type, wave_col = 'nir', 1
    elif cols.str.contains('VIS|KVIS').any():
        data_type, wave_col = 'vis', 2
    else:
        data_type, wave_col = '', 1

    # Extract wavelength array
    wave = df.iloc[:, -wave_col].to_numpy(dtype=int)

    # Compute spatial dimensions from last sample column
    sample_cols = cols[is_sample]
    if sample_cols.size:
        len_x, len_y = _extract_dims(sample_cols[-1])
    else:
        # fallback to raw if no sample columns
        raw_cols = cols[is_raw]
        len_x, len_y = _extract_dims(raw_cols[-1])

    # Filter out drop columns
    df = df.loc[:, ~is_drop]

    # Select data array
    data_cols = cols[is_sample] if is_sample.any() else cols[is_raw]
    data_arr = df[data_cols].to_numpy()

    # Build cube and return
    cube = to_cube(data=data_arr, len_x=len_x, len_y=len_y)
    return DataCube(cube=cube, wavelengths=wave, name=data_type)



import wizard
import numpy as np

# creating some data with the shape 20, 8, 9
some_randome_data = np.zeros(shape=(20, 8, 9))

# create a DataCube
dc = wizard.DataCube(some_randome_data, name='Hello DataCube')

# print dc
print(dc)




"""
_core/datacube_ops.py

.. module:: datacube_ops
   :platform: Unix
   :synopsis: DataCube Operations.

## Module Overview
This module contains operation function for processing datacubes.

## Functions
.. autofunction:: remove_spikes
.. autofunction:: resize
"""
import os
import cv2
import copy
import rembg
import random
import numpy as np
from PIL import Image
from joblib import Parallel, delayed
from scipy.signal import savgol_filter
from scipy.ndimage import gaussian_filter
from skimage.transform import warp


from . import DataCube 
from .._processing.spectral import calculate_modified_z_score, spec_baseline_als
from .._utils.helper import _process_slice, feature_registration, RegistrationError, auto_canny, decompose_homography, normalize_polarity


def remove_spikes(dc: DataCube, threshold: int = 6500, window: int = 5) -> DataCube:
    """
    Remove cosmic spikes from each pixel's spectral data.

    This function computes the modified z-score for each pixel's spectral vector,
    identifies spikes where the score exceeds the threshold, and replaces spike
    values with the local mean within a sliding window along the spectrum.

    Parameters
    ----------
    dc : DataCube
        The input DataCube with shape (v, x, y), where v is the number of spectral bands.
    threshold : int, optional
        Threshold for spike detection via modified z-score, defaults to 6500.
    window : int, optional
        Window size (in spectral channels) for mean replacement of spikes, defaults to 5.

    Returns
    -------
    DataCube
        A new DataCube instance with spikes removed per-pixel.

    Raises
    ------
    ValueError
        If `window` is not in the range [1, number of spectral bands].

    Notes
    -----
    - The original DataCube is not modified in place; es wird eine Kopie zurückgegeben.
    - Die Modifizierte z-Score-Berechnung erwartet Input mit Form (n_samples, n_features).
    - Parallelisierung beschleunigt die Einzelpixel-Bearbeitung.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read("example.fsm")
    >>> dc.remove_spikes(threshold=6500, window=5)
    """
    v, x, y = dc.cube.shape
    if not (1 <= window <= v):
        raise ValueError(f"window must be between 1 and {v}, got {window}")

    # reshape to (n_pixels, v)
    n_pixels = x * y
    flat_cube = dc.cube.reshape(v, n_pixels).T  # shape: (n_pixels, v)

    # Berechne pro-pixel modifizierten z-score
    z_scores = calculate_modified_z_score(flat_cube)  # (n_pixels, v)
    spikes = np.abs(z_scores) > threshold
    flat_out = flat_cube.copy()

    # Parallel auf jedes Pixel anwenden
    results = Parallel(n_jobs=-1)(
        delayed(_process_slice)(flat_out, spikes, idx, window)
        for idx in range(n_pixels)
    )
    for idx, spec in results:
        flat_out[idx] = spec

    # zurück in (v, x, y) formen
    clean_cube = flat_out.T.reshape(v, x, y)

    # Kopie des DataCube mit dem bereinigten Cube

    dc.set_cube(clean_cube)
    return dc


def remove_background(dc: DataCube, threshold: int = 50, style: str = 'dark') -> DataCube:
    """
    Remove background from images in a DataCube.

    Uses an external algorithm (rembg). The first image in the DataCube
    is processed to generate a mask, which is then applied to all images
    to remove the background.

    Parameters
    ----------
    dc : DataCube
        DataCube containing the image stack.
    threshold : int, optional
        Threshold value to define the background from the alpha mask,
        defaults to 50. Pixels with alpha < threshold are considered background.
    style : str, optional
        Style of background removal, 'dark' or 'bright', defaults to 'dark'.
        If 'dark', background pixels are set to 0.
        If 'bright', background pixels are set to the max value of the cube.y

    Returns
    -------
    DataCube
        DataCube with the background removed.

    Raises
    ------
    ValueError
        If style is not 'dark' or 'bright'.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read("example.fsm")
    >>> dc.remove_background(threshold=50, style='bright') # or style 'dark'
    """
    img = dc.cube[0]
    img = ((img - img.min()) / (img.max() - img.min()) * 255).astype('uint8')
    img = Image.fromarray(img)
    img_removed_bg = rembg.remove(img)
    mask = np.array(img_removed_bg.getchannel('A'))

    cube = dc.cube.copy()
    if style == 'dark':
        cube[:, mask < threshold] = 0
    elif style == 'bright':
        cube[:, mask < threshold] = dc.cube.max()
    else:
        raise ValueError("Type must be 'dark' or 'bright'")
    dc.set_cube(cube)
    return dc


def resize(dc: DataCube, x_new: int, y_new: int, interpolation: str = 'linear') -> None:
    """
    Resize the DataCube to new x and y dimensions.

    Resizes each 2D slice (x, y) of the DataCube using the specified
    interpolation method.

    Interpolation methods:
    - ``linear``: Bilinear interpolation (ideal for enlarging).
    - ``nearest``: Nearest neighbor interpolation (fast but blocky).
    - ``area``: Pixel area interpolation (ideal for downscaling).
    - ``cubic``: Bicubic interpolation (high quality, slower).
    - ``lanczos``: Lanczos interpolation (highest quality, slowest).

    Parameters
    ----------
    dc : DataCube
        The DataCube instance to be resized.
    x_new : int
        The new width (x-dimension).
    y_new : int
        The new height (y-dimension).
    interpolation : str, optional
        Interpolation method, defaults to 'linear'.
        Options: 'linear', 'nearest', 'area', 'cubic', 'lanczos'.

    Returns
    -------
    None
        The DataCube is modified in-place.

    Raises
    ------
    ValueError
        If the interpolation method is not recognized.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read("example.fsm")
    >>> dc.resize(x_new=500, y_new=500)
    """
    mode = None
    shape = dc.cube.shape
    if shape[2] > x_new:
        print('\033[93mx_new is smaller than the existing cube, you will lose information\033[0m')
    if shape[1] > y_new:
        print('\033[93my_new is smaller than the existing cube, you will lose information\033[0m')

    if interpolation == 'linear':
        mode = cv2.INTER_LINEAR
    elif interpolation == 'nearest':
        mode = cv2.INTER_NEAREST
    elif interpolation == 'area':
        mode = cv2.INTER_AREA
    elif interpolation == 'cubic':
        mode = cv2.INTER_CUBIC
    elif interpolation == 'lanczos':
        mode = cv2.INTER_LANCZOS4
    else:
        raise ValueError(f'Interpolation method `{interpolation}` not recognized.')

    _cube = np.empty(shape=(shape[0], y_new, x_new))
    for idx, layer in enumerate(dc.cube):
        _cube[idx] = cv2.resize(layer, (x_new, y_new), interpolation=mode)
    dc.cube = _cube
    dc._set_cube_shape()


def baseline_als(dc: DataCube, lam: float = 1000000, p: float = 0.01, niter: int = 10) -> DataCube:
    """
    Apply Adaptive Smoothness (ALS) baseline correction.

    Iterates through each pixel (spectrum) in the DataCube and subtracts
    the baseline calculated by the `spec_baseline_als` function.

    Parameters
    ----------
    dc : DataCube
        The input DataCube.
    lam : float, optional
        The smoothness parameter for ALS, defaults to 1000000.
        Larger lambda makes the baseline smoother.
    p : float, optional
        The asymmetry parameter for ALS, defaults to 0.01.
        Value between 0 and 1. Controls how much the baseline is pushed
        towards the data (0 for minimal, 1 for maximal).
    niter : int, optional
        The number of iterations for the ALS algorithm, defaults to 10.

    Returns
    -------
    DataCube
        The DataCube with baseline correction applied.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read("example.fsm")
    >>> dc.baseline_als(lam=1e6, p=.001, niter=10)
    """
    for x in range(dc.shape[1]):
        for y in range(dc.shape[2]):
            dc.cube[:, x, y] -= spec_baseline_als(
                spectrum=dc.cube[:, x, y],
                lam=lam,
                p=p,
                niter=niter
            )
    return dc


def merge_cubes(dc1: DataCube, dc2: DataCube, register: bool = False) -> DataCube:
    """
    Merge two DataCubes into a single DataCube, with optional registration.

    If both datacubes are already registered and the `register` flag is True,
    the function will sample up to 10 random spectral layers from dc2, attempt to
    register each to the first layer of dc1, choose the transform with the lowest
    mean-squared-error alignment, then apply that best transform to all layers of dc2
    before merging.

    Parameters
    ----------
    dc1 : DataCube
        The first DataCube (used as reference).
    dc2 : DataCube
        The second DataCube to be merged into the first.
    register : bool, optional
        If True (default), registration will be attempted if both cubes are marked as registered.

    Returns
    -------
    DataCube
        A new DataCube containing merged spatial and spectral data.

    Raises
    ------
    NotImplementedError
        If the cubes have mismatched spatial dimensions and cannot be merged,
        or if wavelengths overlap without being purely indices.

    Examples
    --------
    >>> import wizard
    >>> dc_a = wizard.read('example.fsm')
    >>> dc_b = wizard.read('another_file.csv')
    >>> dc_a.merge_cubes(dc_b)
    """
    c1 = dc1.cube
    c2 = dc2.cube
    wave1 = dc1.wavelengths
    wave2 = dc2.wavelengths

    # Optional registration step with sampling
    if register and getattr(dc1, 'registered', False) and getattr(dc2, 'registered', False):
        print("Both datacubes registered. Sampling layers for alignment...")
        ref_img = c1[0]
        num_layers = c2.shape[0]
        sample_indices = random.sample(range(num_layers), min(10, num_layers))

        best_score = np.inf
        best_transform = None
        # Try to register sampled layers and pick best
        for idx in sample_indices:
            try:
                aligned_slice, transform = feature_registration(ref_img, c2[idx])
                # Compute alignment quality (mean squared error)
                mse = np.mean((ref_img - aligned_slice)**2)
                if mse < best_score:
                    best_score = mse
                    best_transform = transform
            except RegistrationError as e:
                print(f"Registration of sampled layer {idx} failed: {e}")

        if best_transform is not None:
            # Apply best transform to all layers of dc2
            for i in range(num_layers):
                try:
                    c2[i] = warp(c2[i], inverse_map=best_transform.inverse, preserve_range=True)
                except Exception as e:
                    print(f"Failed to apply best transform to layer {i}: {e}")
        else:
            print("No successful sampled registration. Skipping registration.")

    # Spatial size check
    if c1.shape[1:] == c2.shape[1:]:
        c3 = np.concatenate([c1, c2], axis=0)
    else:
        raise NotImplementedError(
            'Sorry - this function can only merge cubes with the same spatial dimensions.'
        )

    # Handle wavelength merge
    if set(wave1) & set(wave2):
        # If wavelengths are index-based, just concatenate indices
        if set(wave1) <= set(range(c1.shape[0])) and set(wave2) <= set(range(c2.shape[0])):
            wave3 = list(range(c1.shape[0] + c2.shape[0]))
        else:
            raise NotImplementedError(
                'Sorry - your wavelengths overlap and are not purely index-based.'
            )
    else:
        wave3 = np.concatenate((wave1, wave2))

    # Create new merged DataCube (modify dc1 in-place)
    dc1.set_cube(c3)
    dc1.set_wavelengths(wave3)

    return dc1


def inverse(dc: DataCube) -> DataCube:
    """
    Invert the DataCube values.

    This operation is useful for converting between transmission and
    reflectance data, or similar inversions. The formula applied is:
    `tmp = cube * -1`
    `tmp += -tmp.min()`
    The data type of the cube is preserved if it's 'uint16' or 'uint8'
    after temporary conversion to 'float16' for calculation.

    Parameters
    ----------
    dc : DataCube
        The DataCube to invert.

    Returns
    -------
    DataCube
        The DataCube with inverted values.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read('example.fsm')
    >>> dc.inverse()
    """
    dtype = dc.cube.dtype
    if dtype == np.uint16 or dtype == np.uint8:  # Use np types for comparison
        cube = dc.cube.astype(np.float16)
    else:
        cube = dc.cube.copy()

    tmp = cube
    tmp *= -1
    tmp += -tmp.min()

    dc.set_cube(tmp.astype(dtype))
    return dc


def register_layers_simple(dc: DataCube, max_features: int = 5000, match_percent: float = 0.1) -> DataCube:
    """
    Align images within a DataCube using simple feature-based registration.

    Each layer in the DataCube is aligned to the first layer (index 0)
    using ORB feature detection and homography estimation via `_feature_registration`.

    Parameters
    ----------
    dc : DataCube
        The DataCube whose layers are to be registered.
    max_features : int, optional
        Maximum number of keypoint regions to detect, defaults to 5000.
    match_percent : float, optional
        Percentage of keypoint matches to consider for homography,
        defaults to 0.1 (10%).

    Returns
    -------
    DataCube
        The DataCube with layers registered.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read('example.fsm')
    >>> dc.register_layers_simple()
    """
    o_img = dc.cube[0, :, :]
    for i in range(dc.cube.shape[0]):
        if i > 0:
            a_img = copy.deepcopy(dc.cube[i, :, :])
            try:
                _, h = feature_registration(
                    o_img=o_img, a_img=a_img,
                    max_features=max_features, match_percent=match_percent
                )
                height, width = o_img.shape
                aligned_img = cv2.warpPerspective(a_img, h, (width, height))
                dc.cube[i, :, :] = aligned_img
            except RegistrationError as e:
                print(f"Warning: Could not register layer {i} in simple registration: {e}")
                pass
    dc.registered = True
    return dc


def remove_vignetting_poly(dc: DataCube, axis: int = 1, slice_params: dict = None) -> DataCube:
    """
    Remove vignetting using polynomial fitting along a specified axis.

    Calculates the mean along the specified axis (1 for rows, 2 for columns)
    for each spectral layer, fits a polynomial to this mean profile
    (after Savitzky-Golay smoothing), and subtracts this fitted profile
    from the corresponding rows/columns of the layer to correct for vignetting.

    Parameters
    ----------
    dc : DataCube
        The DataCube instance to process.
    axis : int, optional
        The axis along which to calculate the mean and apply correction.
        1 for correcting along rows (profile used for columns),
        2 for correcting along columns (profile used for rows). Defaults to 1.
    slice_params : dict, optional
        Dictionary for slicing behavior before mean calculation.
        Keys: ``"start"`` (int), ``"end"`` (int), ``"step"`` (int).
        Defaults to full slice with step 1.

    Returns
    -------
    DataCube
        The processed DataCube with vignetting removed.

    Raises
    ------
    ValueError
        If the DataCube is empty or axis is not 1 or 2.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read('example.fsm')
    >>> params = {"start":25, "end":50}
    >>> dc.remove_vignetting_poly(slice_params=params, axis=2)
    """
    if dc.cube is None:
        raise ValueError("The DataCube is empty. Please provide a valid cube.")

    if slice_params is None:
        slice_params = {"start": None, "end": None, "step": 1}
    start = slice_params.get("start", None)
    end = slice_params.get("end", None)
    step = slice_params.get("step", 1)

    if axis == 1:
        summed_data = np.mean(dc.cube[:, :, start:end:step], axis=2)
    elif axis == 2:
        summed_data = np.mean(dc.cube[:, start:end:step, :], axis=1)
    else:
        raise ValueError('Axis can only be 1 or 2.')

    corrected_cube = dc.cube.copy().astype(np.float32)

    for i, layer_profile in enumerate(summed_data):
        smoothed_layer_profile = savgol_filter(layer_profile, window_length=71, polyorder=1)
        if axis == 1:
            for j_col in range(dc.cube.shape[2]):
                corrected_cube[i, :, j_col] -= smoothed_layer_profile
        elif axis == 2:
            for j_row in range(dc.cube.shape[1]):
                corrected_cube[i, j_row, :] -= smoothed_layer_profile

    dc.set_cube(corrected_cube)
    return dc


def normalize(dc: DataCube) -> DataCube:
    """
    Normalize spectral information in the data cube to the range [0, 1].

    For each 2D spatial layer in the DataCube, the normalization is performed by:
    `layer = (layer - min_in_layer) / (max_in_layer - min_in_layer)`
    This scales the intensity values of each layer independently across its
    spatial dimensions.

    Parameters
    ----------
    dc : DataCube
        The DataCube instance to normalize.

    Returns
    -------
    DataCube
        The normalized DataCube.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read('example.fsm')
    >>> dc.normalize()
    """
    cube = dc.cube.astype(np.float32)
    min_vals = cube.min(axis=(1, 2), keepdims=True)
    max_vals = cube.max(axis=(1, 2), keepdims=True)

    range_vals = max_vals - min_vals
    range_vals[range_vals == 0] = 1

    cube = (cube - min_vals) / range_vals
    dc.set_cube(cube)
    return dc


def register_layers_best(
        dc: DataCube,
        ref_layer: int = 0,
        max_features: int = 5000,
        match_percent: float = 0.1,
        rot_thresh: float = 20.0,
        scale_thresh: float = 1.1
) -> DataCube:
    """
    Align DataCube layers with robust registration.

    Aligns each slice of `dc.cube` to a reference layer. It uses
    feature-based registration primarily. If feature-based registration
    yields a degenerate homography (based on rotation and scale thresholds)
    or fails, it falls back to Canny-based edge registration.
    Failed alignments are retried once.

    Parameters
    ----------
    dc : DataCube
        The DataCube to process.
    ref_layer : int, optional
        Index of the reference layer, defaults to 0.
    max_features : int, optional
        Maximum features for ORB, defaults to 5000.
    match_percent : float, optional
        Match percentage for ORB, defaults to 0.1.
    rot_thresh : float, optional
        Rotation threshold (degrees) for homography validation,
        defaults to 20.0.
    scale_thresh : float, optional
        Scale threshold for homography validation, defaults to 1.1.
        Checks if max_scale <= scale_thresh and min_scale >= 1/scale_thresh.

    Returns
    -------
    DataCube
        The DataCube with aligned layers.

    Raises
    ------
    RuntimeError
        If alignment fails for a layer after retry.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read('example.fsm')
    >>> dc.register_layers_best()
    """
    aligned_indices = {ref_layer}
    waitlist = set()
    n_layers, H_dim, W_dim = dc.cube.shape

    def try_align(layer_idx: int, current_aligned_indices: set) -> bool:
        # nonlocal dc
        a_img = dc.cube[layer_idx]
        best_alignment_img = None

        for ref_idx in current_aligned_indices:
            try:
                o_img = dc.cube[ref_idx]
                aligned_img_feat, H_ij = feature_registration(
                    o_img, a_img, max_features, match_percent
                )
                angle, S = decompose_homography(H_ij)
                s_ok = (S.max() <= scale_thresh and S.min() >= 1 / scale_thresh)
                if abs(angle) <= rot_thresh and s_ok:
                    print(f"[Layer {layer_idx}] aligned to {ref_idx}: θ={angle:.1f}°, S={S.round(3)}")
                    best_alignment_img = aligned_img_feat
                    break
                else:
                    print(f"[Layer {layer_idx}] reject vs {ref_idx}: θ={angle:.1f}°, S={S.round(3)}")
            except RegistrationError as e:
                print(f"[Layer {layer_idx}] registration to {ref_idx} failed: {e}")

        if best_alignment_img is None:
            print(f"[Layer {layer_idx}] edge-map fallback to reference layer {ref_layer}")
            try:
                ref_img_for_edge = normalize_polarity(dc.cube[ref_layer])
                tgt_img_for_edge = normalize_polarity(a_img)
                edges_ref = auto_canny(ref_img_for_edge)
                edges_tgt = auto_canny(tgt_img_for_edge)
                aligned_img_edge, H_e = feature_registration(
                    edges_ref.astype(float), edges_tgt.astype(float),
                    max_features, match_percent
                )
                h_orig, w_orig = a_img.shape
                best_alignment_img = cv2.warpPerspective(a_img, H_e, (w_orig, h_orig), flags=cv2.INTER_LINEAR)
                angle_e, S_e = decompose_homography(H_e)
                print(f"[Layer {layer_idx}] edges (vs layer {ref_layer}): θ={angle_e:.1f}°, S={S_e.round(3)}")
            except RegistrationError as e:
                print(f"[Layer {layer_idx}] edge registration failed: {e}")
            except Exception as e:
                print(f"[Layer {layer_idx}] unexpected error in edge registration: {e}")

        if best_alignment_img is not None:
            dc.cube[layer_idx] = best_alignment_img
            return True
        return False

    for i in range(n_layers):
        if i == ref_layer:
            continue
        if try_align(i, aligned_indices.copy()):
            aligned_indices.add(i)
        else:
            waitlist.add(i)

    if waitlist:
        print(f"\nRetrying layers: {list(waitlist)}\n")
        for i in list(waitlist):
            if try_align(i, aligned_indices.copy()):
                aligned_indices.add(i)
                waitlist.remove(i)
            else:
                raise RuntimeError(f"Layer {i}: alignment failed after retry.")
    dc.registered = True
    return dc


def remove_vignetting(dc: DataCube, sigma: float = 50, clip: bool = True, epsilon: float = 1e-6) -> DataCube:
    """
    Remove vignetting from a hyperspectral DataCube.

    Corrects vignetting in each spectral band by estimating a smooth
    background using Gaussian blur and then performing flat-field correction.
    The background is normalized by its mean before correction.

    Parameters
    ----------
    dc : DataCube
        The input DataCube (bands, height, width).
    sigma : float, optional
        Standard deviation for Gaussian blur, controlling smoothness.
        Larger sigma means coarser background estimation. Defaults to 50.
    clip : bool, optional
        If True and the original DataCube dtype is integer,
        clip output values to the valid range of that integer type.
        Defaults to True.
    epsilon : float, optional
        A small constant to add to the background before division
        to prevent division by zero errors. Defaults to 1e-6.

    Returns
    -------
    DataCube
        The DataCube with vignetting corrected. The output cube has the
        same shape and dtype as the input.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read('example.fsm')
    >>> dc.remove_vignetting()
    """
    corrected_cube = np.empty_like(dc.cube)
    orig_dtype = dc.cube.dtype
    is_int = np.issubdtype(orig_dtype, np.integer)

    for i in range(dc.cube.shape[0]):
        band = dc.cube[i].astype(np.float64)
        background = gaussian_filter(band, sigma=sigma)
        background = np.maximum(background, epsilon)
        background_mean = background.mean()
        if background_mean > epsilon:
            background /= background_mean
        else:
            background = np.ones_like(background, dtype=np.float64)
        corrected_band = band / background
        if is_int:
            info = np.iinfo(orig_dtype)
            corrected_band = np.round(corrected_band)
            corrected_band = np.clip(corrected_band, info.min, info.max)
        corrected_cube[i] = corrected_band.astype(orig_dtype)
    dc.set_cube(corrected_cube)
    return dc


def upscale_datacube_edsr(dc: DataCube, scale: int, model_path: str):
    """
    Upscale the spatial dimensions of a DataCube using the EDSR super-resolution model.

    This function applies the EDSR (Enhanced Deep Super-Resolution) model from OpenCV’s
    dnn_superres module to each spectral slice of the given DataCube. Since EDSR
    expects a 3-channel input, each single-band slice is duplicated across three
    channels before processing. The output slices are then reassembled into a new
    DataCube with upscaled spatial dimensions.

    Parameters
    ----------
    dc : DataCube
        An instance of the DataCube class. Must have attributes:
        - datacube.cube: numpy array of shape (v, x, y)
        - datacube.wavelength: list of length v
    scale : int
        The upscaling factor (e.g., 2, 3, or 4) supported by the EDSR model.
    model_path : str
        Filesystem path to the pretrained EDSR `.pb` model file
        (e.g., `"EDSR_x4.pb"`).

    Returns
    -------
    DataCube
        A new DataCube instance whose `.cube` has shape
        (v, x * scale, y * scale) and the same `.wavelength` list.

    Raises
    ------
    FileNotFoundError
        If the specified `model_path` does not exist or is unreadable.
    cv2.error
        If OpenCV fails to load the model or perform upsampling.
    ValueError
        If the `scale` is not one of the factors supported by the loaded model.

    Notes
    -----
    - Requires `opencv-contrib-python>=4.3.0`.
    - Processing is done slice-by-slice and may be slow for large cubes.
    - Memory usage grows by approximately `scale^2`.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read("hyperspectral.fsm")
    >>> dc.upscale_datacube_edsr(scale=4, model_path="EDSR_x4.pb")
    >>> print(dc.cube.shape)
    (v, x*4, y*4)
    """
    # Create EDSR super-res object
    if not os.path.isfile(model_path):
        raise FileNotFoundError(f"Model file not found: {model_path}")
    sr = cv2.dnn_superres.DnnSuperResImpl_create()
    sr.readModel(model_path)
    sr.setModel("edsr", scale)

    v, x, y = dc.shape
    new_x, new_y = x * scale, y * scale
    up_cube = np.empty((v, new_x, new_y), dtype=dc.cube.dtype)

    for i in range(v):
        # Extract single-band slice
        band = dc.cube[i, :, :]

        # Convert to 3-channel BGR by stacking
        bgr = cv2.merge([band, band, band])

        # Upsample
        up_bgr = sr.upsample(bgr)

        # Take one channel (all channels are identical)
        up_band = up_bgr[:, :, 0]

        up_cube[i, :, :] = up_band

    # Build new DataCube
    dc.set_cube(up_cube)
    return dc


def upscale_datacube_espcn(dc: DataCube, scale: int, model_path: str):
    """
    Upscale the spatial dimensions of a DataCube using the ESPCN super-resolution model.

    This function applies the ESPCN (Efficient Sub-Pixel Convolutional Neural Network)
    model from OpenCV’s dnn_superres module to each spectral slice of the given DataCube.
    Since ESPCN expects a 3-channel input, each single-band slice is duplicated across
    three channels before processing. The output slices are then reassembled into a new
    DataCube with upscaled spatial dimensions.

    Parameters
    ----------
    dc : DataCube
        An instance of the DataCube class. Must have attributes:
        - .cube: numpy array of shape (v, x, y)
        - .wavelength: list of length v
    scale : int
        The upscaling factor (2, 3, or 4) supported by the ESPCN model.
    model_path : str
        Filesystem path to the pretrained ESPCN `.pb` model file
        (e.g., "ESPCN_x3.pb").

    Returns
    -------
    DataCube
        A new DataCube instance whose `.cube` has shape
        (v, x * scale, y * scale) and the same `.wavelength` list.

    Raises
    ------
    FileNotFoundError
        If the specified `model_path` does not exist.
    ValueError
        If `scale` is not in {2, 3, 4}.
    cv2.error
        If OpenCV fails to load the model or perform upsampling.

    Notes
    -----
    - Requires `opencv-contrib-python>=4.3.0`.
    - Processing is done slice-by-slice and may be slow for large cubes.
    - Memory usage grows by approximately `scale**2`.

    Examples
    --------
    >>> import wizard
    >>> dc = wizard.read("test.fsm")
    >>> dc.upscale_datacube_espcn(scale=3, model_path="ESPCN_x3.pb")
    >>> print(dc.cube.shape)
    (v, x*3, y*3)
    """
    if not os.path.isfile(model_path):
        raise FileNotFoundError(f"Model file not found: {model_path}")
    if scale not in (2, 3, 4):
        raise ValueError(f"ESPCN only supports scale factors 2, 3, or 4, got {scale}")

    # Initialize ESPCN super-resolver
    sr = cv2.dnn_superres.DnnSuperResImpl_create()
    sr.readModel(model_path)
    sr.setModel("espcn", scale)

    v, x, y = dc.shape
    new_x, new_y = x * scale, y * scale
    up_cube = np.empty((v, new_x, new_y), dtype=dc.cube.dtype)

    for i in range(v):
        # extract single-band slice
        band = dc.cube[i, :, :]

        # create 3-channel image
        bgr = cv2.merge([band, band, band])

        # run ESPCN upsampling
        up_bgr = sr.upsample(bgr)

        # channels are identical: pick one
        up_band = up_bgr[:, :, 0]
        up_cube[i, :, :] = up_band

    # assemble new DataCube
    dc.set_cube(up_cube)
    return dc


def upscale_datacube_with_reference(dc: DataCube, reference_image: np.ndarray) -> 'DataCube':
    """
    Upscales a DataCube to match the spatial resolution of a reference image using ESRGAN.

    Each spectral band of the DataCube is individually upscaled using the ESRGAN model
    by treating the single band as a pseudo-grayscale image. The resulting DataCube has
    the same spatial resolution as the reference image.

    Parameters
    ----------
    dc : DataCube
        The input DataCube to be upscaled. Expected shape (v, x, y).

    reference_image : np.ndarray
        A high-resolution reference image (e.g., RGB) that defines the target (x, y) resolution.

    Returns
    -------
    DataCube
    A new DataCube with the same number of bands, but upscaled to match the reference image's spatial resolution.

    Raises
    ------
    ValueError
    If the reference image resolution is smaller than the datacube resolution.

    Notes
    -----
    This method uses the ESRGAN model via TensorFlow Hub. The model is applied to each spectral band independently.

    Examples
    --------
    >>> upscaled_cube = upscale_datacube_with_reference(cube, ref_image)
    """
    import tensorflow as tf
    import tensorflow_hub as hub

    # Load ESRGAN model
    model_url = "https://tfhub.dev/captain-pool/esrgan-tf2/1"
    model = hub.load(model_url)

    # Target resolution from reference image
    target_x, target_y = reference_image.shape[:2]

    v, x, y = dc.shape

    if x >= target_x or y >= target_y:
        raise ValueError("Reference image must be larger than the DataCube in spatial dimensions.")

    upscaled_bands = []

    for i in range(v):
        band = dc.cube[i, :, :]

        # Normalize and expand dims for ESRGAN
        img = (band / band.max() * 255).astype(np.uint8)
        img_rgb = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        img_rgb = tf.convert_to_tensor(img_rgb, dtype=tf.float32)

        # Ensure divisible by 4
        imageSize = (tf.convert_to_tensor(img_rgb.shape[:-1]) // 4) * 4
        cropped_image = tf.image.crop_to_bounding_box(
            img_rgb, 0, 0, imageSize[0], imageSize[1]
        )
        preprocessed_image = tf.expand_dims(cropped_image, 0)

        # Run super-resolution
        sr_result = model(preprocessed_image)
        sr_image = tf.squeeze(sr_result).numpy() / 255.0

        # Convert back to single band and resize to match reference
        sr_gray = cv2.cvtColor((sr_image * 255).astype(np.uint8), cv2.COLOR_RGB2GRAY)
        resized_band = cv2.resize(sr_gray.astype(np.float32), (target_y, target_x), interpolation=cv2.INTER_CUBIC)

        upscaled_bands.append(resized_band)

    upscaled_cube_array = np.stack(upscaled_bands, axis=0)

    dc.set_cube(upscaled_cube_array)
    return dc


def upscale_datacube_fsrcnn(dc, scale, model_path):
    """
    Upscale the spatial dimensions of a DataCube using the FSRCNN super-resolution model.

    This function applies the FSRCNN (Fast Super-Resolution Convolutional Neural Network)
    model from OpenCV’s dnn_superres module to each spectral slice of the given DataCube.
    Since FSRCNN expects a 3-channel input, each single-band slice is duplicated across three
    channels before processing. The output slices are then reassembled into a new DataCube
    with upscaled spatial dimensions.

    Parameters
    ----------
    dc : DataCube
        An instance of the DataCube class. Must have attributes:
        - .cube: numpy array of shape (v, x, y)
        - .wavelength: list of length v
    scale : int
        The upscaling factor (2, 3, or 4) supported by the FSRCNN model.
    model_path : str
        Filesystem path to the pretrained FSRCNN `.pb` model file
        (e.g., "FSRCNN_x3.pb").

    Returns
    -------
    DataCube
        A new DataCube instance whose `.cube` has shape
        (v, x * scale, y * scale) and the same `.wavelength` list.

    Raises
    ------
    FileNotFoundError
        If the specified `model_path` does not exist.
    ValueError
        If `scale` is not in {2, 3, 4}.
    cv2.error
        If OpenCV fails to load the model or perform upsampling.

    Notes
    -----
    - Requires `opencv-contrib-python>=4.3.0`.
    - Processing is done slice-by-slice and may be slow for large cubes.
    - Memory usage grows by approximately `scale**2`.

    Examples
    --------
    >>> dc = wizard.read("hyperspectral.npy", wavelengths=[...])
    >>> up_dc = upscale_datacube_fsrcnn(dc, scale=3, model_path="FSRCNN_x3.pb")
    >>> print(up_dc.cube.shape)
    (v, x*3, y*3)
    """
    # Validate model file
    if not os.path.isfile(model_path):
        raise FileNotFoundError(f"Model file not found: {model_path}")
    # Validate supported scale factors
    if scale not in (2, 3, 4):
        raise ValueError(f"FSRCNN only supports scale factors 2, 3, or 4, got {scale}")

    # Initialize FSRCNN super-resolver
    sr = cv2.dnn_superres.DnnSuperResImpl_create()
    sr.readModel(model_path)
    sr.setModel("fsrcnn", scale)

    # Prepare new cube
    v, x, y = dc.cube.shape
    new_x, new_y = x * scale, y * scale
    up_cube = np.empty((v, new_x, new_y), dtype=dc.cube.dtype)

    # Upscale each spectral band
    for i in range(v):
        band = dc.cube[i, :, :]
        bgr = cv2.merge([band, band, band])
        up_bgr = sr.upsample(bgr)
        up_band = up_bgr[:, :, 0]
        up_cube[i, :, :] = up_band

    # Create and return new DataCube
    dc.set_cube(up_cube)
    return dc


def remove_vignette(dc: DataCube, vignette_map: np.ndarray, flip: bool = False) -> DataCube:
    """
    Subtract a vignette pattern from every spectral layer.

    Removes spatial vignetting by subtracting the provided vignette_map
    from each (x, y) layer in the data cube. If flip=True, the vignette
    pattern is inverted (dark center → bright center) before subtraction.

    Parameters
    ----------
    dc: DataCube
        An instance of the DataCube class. Must have attributes:
        - .cube: numpy array of shape (v, x, y)
        - .wavelength: list of length v
    vignette_map : np.ndarray
        2D array of shape (x, y) representing the vignette intensity to subtract.
        Values should be on the same scale as the cube’s pixel intensities.
    flip : bool, default=False
        If True, invert the vignette_map before subtraction.

    Returns
    -------
    None
        Modifies the DataCube.cube in-place.

    Raises
    ------
    ValueError
        If vignette_map.shape does not match the spatial dimensions of the cube.

    Notes
    -----
    - After subtraction, any negative values in the cube are clipped to zero.
    - Assumes cube and vignette_map share the same intensity scale.

    Examples
    --------
    >>> dc = DataCube(cube=np.random.rand(10, 256, 256), wavelength=list(np.linspace(400,700,10)))
    >>> # Remove standard vignette (darker at edges, brighter center)
    >>> dc.remove_vignette(vignette_map=my_vignette_image)
    >>> # Remove flipped vignette (darker center, brighter edges)
    >>> dc.remove_vignette(vignette_map=my_vignette_image, flip=True)
    """
    # Check that the vignette map matches spatial dims
    if vignette_map.shape != dc.cube.shape[1:]:
        raise ValueError(
            f"vignette_map shape {vignette_map.shape} does not match cube spatial shape {dc.cube.shape[1:]}"
        )

    cube = dc.cube.copy()

    # Optionally invert the vignette pattern
    if flip:
        vignette_map = vignette_map.max() - vignette_map

    # Subtract vignette from each layer
    # Using broadcasting: vignette_map has shape (x, y), expand to (1, x, y)
    cube -= vignette_map[np.newaxis, :, :]

    # Clip negative values to zero
    np.clip(cube, a_min=0, a_max=None, out=cube)

    dc.set_cube(cube)

    return dc



"""
_utils/__init__.py
===================

.. module:: __init__
   :platform: Unix
   :synopsis: Initialization module for the hsi-wizard package.

Module Overview
---------------

This module initializes the _utils package, making utility functions
available for use throughout the hsi-wizard package.

Functions
---------

.. autofunction:: find_nex_greater_wave
.. autofunction:: find_nex_smaller_wave

"""
from .helper import find_nex_greater_wave, find_nex_smaller_wave



"""
_utils/_loader/__init__.py
============================

.. module:: __init__.py
   :platform: Unix
   :synopsis: Initialization for the loader functions in the _utils._loader package.

Module Overview
---------------

This module initializes loader and writer functions for various file types. It allows dynamic registration of loaders based on file extensions.

Functions
---------
.. autofunction:: register_loader
.. autofunction:: read
.. autofunction:: load_all_loaders

"""

import pathlib
import importlib
import os

# Dictionary to register loaders based on file extensions
LOADER_REGISTRY = {}

# Populate __all__ to control the public API
__all__ = ['read']


def register_loader(extension, function_name):
    """
    Register a new loader for a specific file extension.

    :param extension: File extension (e.g., '.csv').
    :type extension: str
    :param function_name: Loader function to be registered (e.g., read_csv).
    :type function_name: callable
    """
    LOADER_REGISTRY[extension] = function_name


def read(path: str, datatype: str = 'auto', **kwargs):
    """
    Read data from files of various types and return a DataCube object.

    :param path: Path to the data file.
    :type path: str
    :param datatype: Data type of the file (e.g., '.csv', '.xlsx'). If 'auto', the file extension is inferred from the path.
    :type datatype: str
    :param kwargs: Additional keyword arguments passed to the loader function.
    :return: DataCube object containing the imported data.
    :rtype: DataCube
    :raises NotImplementedError: If no loader is registered for the specified file type.
    """
    if datatype == 'auto':
        if os.path.isdir(path):
            suffix = '.folder'
        elif path.endswith('tiff') or path.endswith('jpg') or path.endswith('png'):
            suffix = '.image'
        else:
            suffix = pathlib.Path(path).suffix
    else:
        suffix = datatype

    # Get the loader function based on the file extension
    loader_function = LOADER_REGISTRY.get(suffix)

    if loader_function:
        return loader_function(path, **kwargs)
    else:
        raise NotImplementedError(f'No loader for {suffix} files; please use the custom loader class of the DataCube.')


def load_all_loaders():
    """
    Automatically discover and import loaders from the wizard._utils._loader package.

    This function imports modules corresponding to known file types and registers their associated loading functions.
    """
    loader_modules = [
        "csv",
        "xlsx",
        "tdms",
        "fsm",
        "folder",
        "nrrd",
        "image",
    ]

    for module_name in loader_modules:
        module = importlib.import_module(f'wizard._utils._loader.{module_name}')
        for attr_name in dir(module):
            if attr_name.startswith('_read_'):
                # Assuming the function name is read_csv, read_xlsx, etc.
                extension = '.' + attr_name.split('_')[-1]  # e.g., 'read_csv' -> '.csv'
                loader_function = getattr(module, attr_name)
                register_loader(extension, loader_function)


# Load all loaders dynamically
load_all_loaders()



"""
_utils/tracker.py
=================

.. module:: tracker
   :platform: Unix
   :synopsis: Tracker functions for monitoring DataCube changes.

Module Overview
---------------

This module contains functions to track changes made to DataCube instances.

Classes
-------

.. autoclass:: TrackExecutionMeta
   :members:
   :undoc-members:
   :show-inheritance:
"""

excluded = ['stop_recording', 'save_template', '_clean_data', '_map_args_to_kwargs', 'execute_template']


class TrackExecutionMeta(type):
    """
    Metaclass for tracking method executions in classes that use it.

    Automatically wraps dynamic methods (marked with `__is_dynamic__ = True`)
    to log their calls when tracking is enabled. Used for debugging, auditing,
    or analyzing usage patterns of dynamic methods.

    Attributes
    ----------
    recording : bool
        Indicates whether method tracking is currently active.

    recorded_methods : list of tuple
        Stores a list of tuples (method_name, args, kwargs) representing
        recorded method calls.

    Methods
    -------
    start_recording()
        Enables method tracking and clears previously recorded data.

    stop_recording()
        Disables method tracking.

    record_method(func)
        Decorator to wrap and conditionally record dynamic method calls.
    """

    recording = False
    recorded_methods = []

    def __new__(cls, name, bases, dct):
        """
        Creates a new class, wrapping its methods for tracking if applicable.

        Wraps all callable attributes except those explicitly excluded or not marked dynamic.

        Parameters
        ----------
        name : str
            Name of the class being created.

        bases : tuple
            Base classes of the new class.

        dct : dict
            Dictionary of the class's attributes and methods.

        Returns
        -------
        type
            A new class with method execution tracking wrappers.
        """
        for key, value in dct.items():
            if callable(value) and key != 'execute_template':
                dct[key] = cls.record_method(value)
        return super().__new__(cls, name, bases, dct)

    @staticmethod
    def record_method(func):
        """
        Wraps a method to record its execution if tracking is enabled.

        Only records calls to methods explicitly marked as dynamic
        using the `__is_dynamic__ = True` attribute.

        Parameters
        ----------
        func : callable
            The method to be wrapped.

        Returns
        -------
        callable
            The wrapped method that conditionally records executions.

        Notes
        -----
        This does not alter the method's original behavior, only adds tracking.
        """
        def wrapper(*args, **kwargs):
            if TrackExecutionMeta.recording:
                if getattr(func, '__is_dynamic__', False):
                    print(f"Tracking dynamic method: {func.__name__}")
                    TrackExecutionMeta.recorded_methods.append(
                        (func.__name__, args, kwargs))
            return func(*args, **kwargs)
        return wrapper

    @staticmethod
    def start_recording():
        """
        Enables method tracking and resets previously recorded calls.

        Notes
        -----
            Use this before any tracked operation to capture method calls.
        """
        TrackExecutionMeta.recording = True
        TrackExecutionMeta.recorded_methods = []

    @staticmethod
    def stop_recording():
        """
        Disables method tracking.

        Notes
        -----
            Use this to halt tracking when the desired operations are complete.
        """
        TrackExecutionMeta.recording = False



"""
_core/__init__.py
<<<<<<< HEAD
=================

.. module:: __init__
   :platform: Unix
   :synopsis: Initialization of the core package for `hsi-wizard`.

Module Overview
---------------

This module initializes the core functionality of the `hsi-wizard` package, including attaching
dynamic methods from `datacube_ops` to the `DataCube` class. It ensures that operations are
dynamically bound to `DataCube` during runtime, and tracks execution using the `TrackExecutionMeta`
class from the `wizard._utils.tracker` module.

Functions
---------

.. autofunction:: attach_datacube_operations
.. autofunction:: func_as_method

"""

import multiprocessing
from functools import wraps

from .datacube import DataCube
from wizard._utils.tracker import TrackExecutionMeta

__all__ = ['DataCube']


def func_as_method(func):
    """
    Converts a standalone function into a method that can be dynamically attached to a class.

    This function ensures that the method retains the original function's name and signature
    by using the `wraps` decorator.

    Parameters
    ----------
    func : function
        The standalone function to be converted into a method.

    Returns
    -------
    function
        The wrapped function as a method, with a flag indicating it is dynamic.
    """
    @wraps(func)
    def method(self, *args, **kwargs):
        return func(self, *args, **kwargs)

    # Mark the function as dynamic for tracking purposes
    method.__is_dynamic__ = True
    return method


_dynamic_methods_attached = False  # Global flag to prevent repeated attachments


def attach_datacube_operations():
    """
    Dynamically attach functions from `datacube_ops` to the `DataCube` class as methods.

    This function imports `datacube_ops` and attaches all callable operations as methods to
    the `DataCube` class. Each method is wrapped with the `TrackExecutionMeta.record_method`
    decorator to enable execution tracking.

    This operation is only performed once, controlled by the `_dynamic_methods_attached` flag,
    and only in the main process to avoid issues in multiprocessing contexts.

    Notes
    -----
    - Functions from `datacube_ops` are converted into methods using `func_as_method`.
    - Execution tracking is applied to all dynamic methods using `TrackExecutionMeta`.

    Returns
    -------
    None
    """
    from . import datacube_ops  # prevent circular import errors
    global _dynamic_methods_attached
    if not _dynamic_methods_attached:
        # Ensure we're in the main process, not a child process
        if multiprocessing.current_process().name == "MainProcess":
            for name in dir(datacube_ops):
                func = getattr(datacube_ops, name)
                if callable(func):
                    # Wrap the method with the tracking decorator before attaching
                    wrapped_func = TrackExecutionMeta.record_method(func_as_method(func))
                    setattr(DataCube, name, wrapped_func)  # Attach the wrapped function
            _dynamic_methods_attached = True  # Set flag after the first call


# Attach methods to the DataCube class on import
attach_datacube_operations()



"""
_utils/_loader/_helper.py
===========================

.. module:: _helper
   :platform: Unix
   :synopsis: Utility functions for file operations and data transformations.

Module Overview
---------------

This module provides helper functions for various file operations, including transforming data arrays,
retrieving files by extension, and converting paths to absolute form.

Functions
---------

.. autofunction:: to_cube
.. autofunction:: get_files_by_extension
.. autofunction:: make_path_absolute

"""

import os
import glob
import numpy as np


def to_cube(data: np.array, len_x: int, len_y: int) -> np.array:
    """
    Transform a 1D numpy array into a 3D data cube-like array.

    The transformation uses Fortran-like index ordering, which means the
    data is reshaped according to Fortran order.

    :param data: 1D array of data in Fortran order.
    :type data: np.ndarray
    :param len_x: The length of the data cube along the x-axis (pixel size x).
    :type len_x: int
    :param len_y: The length of the data cube along the y-axis (pixel size y).
    :type len_y: int
    :return: Transformed 3D data cube.
    :rtype: np.ndarray
    """
    return data.reshape(-1, len_x, len_y, order='F')


def get_files_by_extension(path: str, extension: str) -> list:
    """
    Retrieve a sorted list of filenames with a specified extension from a directory.

    :param path: Directory path to search for files.
    :type path: str
    :param extension: File extension to filter by (e.g., `.csv`).
    :type extension: str
    :return: Sorted list of filenames with the specified extension.
    :rtype: list
    """
    if not extension:
        return []

    if not os.path.isdir(path):
        return []

    if not extension.startswith('.'):
        extension = '.' + extension

    return sorted(glob.glob(os.path.join(path, '*' + extension.lower())))


def make_path_absolute(path: str) -> str:
    """
    Convert a relative path to an absolute path if it is not already absolute.

    :param path: Path to the file or directory.
    :type path: str
    :return: Absolute path to the file or directory.
    :rtype: str
    :raises ValueError: If the input path is not a string or is invalid.
    """
    if isinstance(path, str) and path:
        if not os.path.isabs(path):
            path = os.path.abspath(path)
        return path.lower()
    else:
        raise ValueError("Input path must be a non-empty string.")



"""
_processing/spectral.py
========================

.. module:: spectral
   :platform: Unix
   :synopsis: Provides methods for processing data cubes, including spike removal and intensity calculations.

Module Overview
---------------

This module includes functions for processing data cubes. It calculates the modified z-score of a data cube to
identify intensity differences, removes spikes from the data by replacing them with mean values of neighboring data,
and processes data cubes in parallel to enhance performance.

"""

import numpy as np
from scipy.signal import savgol_filter, butter, filtfilt
from scipy import sparse
from scipy.sparse.linalg import spsolve


def smooth_savgol(spectrum, window_length: int = 11, polyorder: int = 2):
    """
    Apply a Savitzky-Golay filter to smooth a given spectrum.

    :param spectrum: Input spectrum to be smoothed.
    :type spectrum: numpy.ndarray
    :param window_length: Length of the filter window, must be a positive odd integer.
    :type window_length: int, optional
    :param polyorder: Order of the polynomial used to fit the samples.
    :type polyorder: int, optional
    :return: Smoothed spectrum.
    :rtype: numpy.ndarray
    """
    return savgol_filter(spectrum, window_length=window_length, polyorder=polyorder)


def smooth_moving_average(spectrum, window_size: int = 5):
    """
    Apply a moving average filter to smooth a given spectrum.

    :param spectrum: Input spectrum to be smoothed.
    :type spectrum: numpy.ndarray
    :param window_size: Size of the moving window.
    :type window_size: int, optional
    :return: Smoothed spectrum.
    :rtype: numpy.ndarray
    """
    return np.convolve(spectrum, np.ones(window_size) / window_size, mode='valid')


def smooth_butter_lowpass(spectrum, cutoff: float = .1, fs: float = 1., order: int = 5):
    """
    Apply a Butterworth low-pass filter to the input spectrum.

    :param spectrum: Input spectrum to be filtered.
    :type spectrum: numpy.ndarray
    :param cutoff: Cutoff frequency of the filter.
    :type cutoff: float, optional
    :param fs: Sampling frequency of the input signal.
    :type fs: float, optional
    :param order: Order of the filter.
    :type order: int, optional
    :return: Filtered spectrum.
    :rtype: numpy.ndarray
    """
    nyquist = 0.5 * fs
    normal_cutoff = cutoff / nyquist
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    return filtfilt(b, a, spectrum)


def spec_baseline_als(spectrum: np.array, lam: float, p: float, niter: int = 10) -> np.array:
    """
    Perform baseline correction using Asymmetric Least Squares Smoothing.

    :param spectrum: Input spectrum to be smoothed.
    :type spectrum: numpy.ndarray
    :param lam: Smoothness parameter, typically between 10^2 and 10^9.
    :type lam: float
    :param p: Asymmetry parameter, usually between 0.001 and 0.1.
    :type p: float
    :param niter: Number of iterations.
    :type niter: int, optional
    :return: Smoothed baseline.
    :rtype: numpy.ndarray
    """
    m = len(spectrum)
    D = sparse.diags([1, -2, 1], [0, 1, 2], shape=(m, m)).tocsc()
    w = np.ones(m)
    for _ in range(niter):
        W = sparse.diags(w, 0).tocsc()
        Z = W + lam * D @ D.T
        z = spsolve(Z, w * spectrum)
        w = p * (spectrum > z) + (1 - p) * (spectrum < z)
    return z


def calculate_modified_z_score(spectrum: np.array):
    """
    Calculate the modified z-score of a data cube by computing the difference in intensity along the first axis.

    :param spectrum: spectrum
    :type spectrum: numpy.ndarray
    :return: Modified z-score representing the difference in intensity along the first axis.
    :rtype: numpy.ndarray
    """
    mean_spectrum = np.mean(spectrum, keepdims=True)  # Compute mean
    return spectrum - mean_spectrum  # Subtract mean from each element


def get_ratio_two_specs(spectrum: np.array, waves: np.array, wave_1: int, wave_2: int) -> np.array:
    """
    Calculate the ratio between two selected wavelengths in the spectrum.

    :param spectrum: Main spectrum from which the wavelengths will be selected.
    :type spectrum: numpy.ndarray
    :param waves: Array of wavelengths for the given spectrum.
    :type waves: numpy.ndarray
    :param wave_1: First selected wavelength.
    :type wave_1: int
    :param wave_2: Second selected wavelength.
    :type wave_2: int
    :return: Ratio of the two selected wavelengths.
    :rtype: numpy.ndarray
    """
    idx_1 = np.searchsorted(waves, wave_1)
    idx_2 = np.searchsorted(waves, wave_2)
    return spectrum[idx_1] / spectrum[idx_2] if idx_1 < len(waves) and idx_2 < len(waves) else -1


def get_sub_tow_specs(spectrum: np.array, waves: np.array, wave_1: int, wave_2: int) -> np.array:
    """
    Subtract two selected wavelengths from a spectrum.

    :param spectrum: Main spectrum from which the wavelengths will be selected.
    :type spectrum: numpy.ndarray
    :param waves: Array of wavelengths for the given spectrum.
    :type waves: numpy.ndarray
    :param wave_1: First selected wavelength.
    :type wave_1: int
    :param wave_2: Second selected wavelength.
    :type wave_2: int
    :return: Difference of the two selected wavelengths.
    :rtype: numpy.ndarray
    """
    idx_1 = np.searchsorted(waves, wave_1)
    idx_2 = np.searchsorted(waves, wave_2)
    return spectrum[idx_1] - spectrum[idx_2] if idx_1 < len(waves) and idx_2 < len(waves) else -1


def signal_to_noise(spectrum: np.array) -> float:
    """
    Calculate the signal-to-noise ratio (SNR) of a given spectrum.

    :param spectrum: 1D array representing the intensity values of the spectrum.
    :type spectrum: numpy.ndarray
    :return: Signal-to-noise ratio (SNR) value in decibels (dB).
    :rtype: float
    """
    mean = np.mean(spectrum)
    std = np.std(spectrum)
    return 20 * np.log10(abs(mean / std)) if std != 0 else 0


def del_leading_zeros(spectrum, idx: int = -1, auto_offset: int = 5):
    """
    Remove leading zeros from a spectrum array.

    :param spectrum: Input spectrum array.
    :type spectrum: numpy.ndarray
    :param idx: Index value for cutting.
    :type idx: int, optional
    :param auto_offset: Number of values to be deleted in auto mode (idx + auto_offset).
    :type auto_offset: int, optional
    :return: Spectrum array with leading zeros removed.
    :rtype: numpy.ndarray
    """
    if idx == -1:
        idx = 0
        for idx, i in enumerate(spectrum):
            if i != 0:
                break
        idx += auto_offset
    return spectrum[idx:]


def del_last_zeros(spectrum, idx: int = -1, auto_offset: int = 5):
    """
    Remove trailing zeros from a spectrum array.

    :param spectrum: Input spectrum array.
    :type spectrum: numpy.ndarray
    :param idx: Index value for cutting.
    :type idx: int, optional
    :param auto_offset: Number of values to be deleted in auto mode (idx + auto_offset).
    :type auto_offset: int, optional
    :return: Spectrum array with trailing zeros removed.
    :rtype: numpy.ndarray
    """
    if idx == -1:
        idx = 0
        for idx, i in enumerate(reversed(spectrum)):
            if i != 0:
                break
        idx += auto_offset
    return spectrum[:idx]



import os
import wizard
from matplotlib import pyplot as plt
from mpmath.libmp import normalize
from rich.progress import track
from pathlib import Path

import numpy as np


fl = '/Users/flx/Documents/Mulitscanner/Data/Images/Mice_Brain/FL_Imager_gold/BrainGold1'
print('FL-Imager:')
fl_dc = wizard.read(fl, normalize=True)
# fl_dc.register_layers()
# fl_dc.inverse()

print('\t- remove vingetting')
fl_dc.remove_vignetting(sigma=100)

print('\t- remove Background')
fl_dc.remove_background()

print('\t- normalize')
fl_dc.normalize()

# print('\tresize')
# fl_dc.resize(x_new=mir_dc.shape[2], y_new=mir_dc.shape[1])
wizard.plotter(fl_dc)

mir = '/Users/flx/Documents/Mulitscanner/Data/Images/Mice_Brain/MIR_gold/Mice Brain gold 2'
print('mir')
mir_dc = wizard.read(mir)
mir_dc.remove_background(style='bright')
# mir_dc.register_layers()
mir_dc.normalize()
mir_dc.set_cube(np.flip(mir_dc.cube, axis=(1, 2)))
# print(mir_dc.shape)
# wizard.plotter(mir_dc)

mir_dc.resize(x_new=fl_dc.shape[2], y_new=fl_dc.shape[1])
mir_dc.set_cube(1-mir_dc.cube)
print(mir_dc.cube.max(), mir_dc.cube.min())
mir_dc.normalize()
wizard.plotter(mir_dc)

print('merge')
fl_dc.merge_cubes(mir_dc)
fl_dc.normalize()
fl_dc.register_layers_best(max_features=10000, scale_thresh=3.)
print(fl_dc.shape)

wizard.plotter(fl_dc)

print('Done')
